//============================================================================
// Name        : ehBinauralMultisourceLocalization.cpp
// Author      : Eui-Hyun Kim (uihyun@gmail.com), Yoshiaki Bando (yoshiaki@kuis.kyoto-u.ac.jp)
// Version     : 2.12.5
// Copyright   : Okuno Laboratory, Kyoto University
// Description :
//============================================================================

#include <iostream>
#include <string>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Map.h>
#include <Source.h>

using namespace std;
using namespace HARK;

class BinauralMultisourceLocalization;

DECLARE_NODE(BinauralMultisourceLocalization);
/*Node
 *
 * @name BinauralMultisourceLocalization
 * @category HARK:BINAURAL+
 * @description Binarual Sound Source Localization based on Generalized Cross-Correlation method weighted by the PHAse Transform (GCC-PHAT)
 *
 * @input_name AUDIO_SPECTRUM
 * @input_type Matrix<complex<float> >
 * @input_description Windowed spectrum data. A row index is channel, and a column index is frequency.
 *
 * @output_name ESTIMATED_DOA
 * @output_type Vector<ObjectRef>
 * @output_description Estimated directions of multisource
 *
 * @parameter_name SOUND_DIFFRACTION_COMPENSATION
 * @parameter_type string
 * @parameter_value FREE_SPACE
 * @parameter_list FREE_SPACE:SPHERICAL_HEAD
 * @parameter_description Compensation for the diffraction of sound waves with multi-path interference caused by contours of a spherical robot head [default: FREE_SPACE].
 *
 * @parameter_name GCCPHAT_THRESHOLD
 * @parameter_type float
 * @parameter_value 0.25
 * @parameter_description Threshold to avoid estimating DOAs of noise and reverberated sources [default: 0.25 (0.0 - 1.0)].
 *
 * @parameter_name MICS_DISTANCE
 * @parameter_type float
 * @parameter_value 17.4
 * @parameter_description Distance between two microphones (cm) [default: 17.4].
 *
 * @parameter_name NOISE_DURATION
 * @parameter_type float
 * @parameter_value 1.0
 * @parameter_description Time duration to be regarded as "noise" from the first frame [default: 1.0 seconds].
 *
 * @parameter_name FREQUENCY_HIGH
 * @parameter_type float
 * @parameter_value 7000.0
 * @parameter_description Upper bound of frequency range where GCC-PHAT is calculated [default: 7000.0 Hz]
 *
 * @parameter_name FREQUENCY_LOW
 * @parameter_type float
 * @parameter_value 60.0
 * @parameter_description Lower bound of frequency range where GCC-PHAT is calculated [default: 60.0 Hz]
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The length in sample between a frame and a previous frame [default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 *
 END*/

class BinauralMultisourceLocalization : public BufferedNode {

  int spectrumInputID;
  int sourcesOutputID;
  int advance;
  int samplingRate;
  float EPSILON;

  float frequency_high;
  float frequency_low;

  // Noise Estimation
  float noiseDuration;
  int numberOfNoiseFrames;
  float* avgNoisePowerSpectrum;
  float* varNoisePowerSpectrum;

  // GCC-PHAT
  float micDistance;
  float gccPhatThreshold;
  String diffractionCompensation;
  float** steeringExponentialR;
  float** steeringExponentialI;

public:
  BinauralMultisourceLocalization(string nodeName, ParameterSet params) : BufferedNode(nodeName, params)
  {
    spectrumInputID = addInput("AUDIO_SPECTRUM");
    sourcesOutputID = addOutput("ESTIMATED_DOA");
    micDistance = dereference_cast<float>(parameters.get("MICS_DISTANCE"));
    gccPhatThreshold = dereference_cast<float>(parameters.get("GCCPHAT_THRESHOLD"));
    diffractionCompensation = object_cast<String>(parameters.get("SOUND_DIFFRACTION_COMPENSATION"));
    advance = dereference_cast<int>(parameters.get("ADVANCE"));
    samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));

    frequency_high = dereference_cast<float>(parameters.get("FREQUENCY_HIGH"));
    frequency_low = dereference_cast<float>(parameters.get("FREQUENCY_LOW"));

    EPSILON = 1e-6;

    steeringExponentialR = 0;
    steeringExponentialI = 0;

    inOrder = true;

    // Noise Estimation
    noiseDuration = dereference_cast<float>(parameters.get("NOISE_DURATION"));
    numberOfNoiseFrames = int(noiseDuration/float(advance)*float(samplingRate));
    if (numberOfNoiseFrames <= 0)
      numberOfNoiseFrames = 1;
    avgNoisePowerSpectrum = 0;
    varNoisePowerSpectrum = 0;

  }

  virtual ~BinauralMultisourceLocalization()
  {
    // release memory of desired signals
    if (steeringExponentialR != 0 && steeringExponentialI != 0) {
      for (int iDegree = -90; iDegree <= +90; iDegree++) {
	delete [] steeringExponentialR[iDegree+90];
	delete [] steeringExponentialI[iDegree+90];
      }
      delete [] steeringExponentialR;
      delete [] steeringExponentialI;
      steeringExponentialR = 0;
      steeringExponentialI = 0;
    }

    if (!avgNoisePowerSpectrum)
      delete [] avgNoisePowerSpectrum;

    if (!varNoisePowerSpectrum)
      delete [] varNoisePowerSpectrum;
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    // set input data
    RCPtr<Matrix<complex<float> > > spectrumInputPtr = getInput(spectrumInputID, count);
    Matrix<complex<float> >& spectrumInput = *spectrumInputPtr;
    int nCols = spectrumInput.ncols()-1;

    // set output data
    RCPtr<Vector<ObjectRef> > sourcesPtr(new Vector<ObjectRef>);
    (*outputs[sourcesOutputID].buffer)[count] = sourcesPtr;

    // The frequency of human voice ranges from about 60 Hz to 7000 Hz
    int frequencyBinFrom  = frequency_low / (samplingRate/2/nCols);
    int frequencyBinTo = frequency_high / (samplingRate/2/nCols);
    int nBinsForAveraging = frequencyBinTo - frequencyBinFrom;

    const int leftMic = 0; const int rightMic = 1;

    // ----------  Averaging noise spectrum : Begin ----------

    // Initialization
    if (count == 0)
      {
	avgNoisePowerSpectrum = new float[nCols];
	for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	  avgNoisePowerSpectrum[iBin] = 0.f;

	varNoisePowerSpectrum = new float[nCols];
	for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	  varNoisePowerSpectrum[iBin] = 0.f;

      } else if (count < numberOfNoiseFrames) {

      ////////////////////////////////////////////////
      // Noise spectrum extraction
      for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	{
	  float CrossCorrelationR = spectrumInput(leftMic,iBin).real()*spectrumInput(rightMic,iBin).real() + spectrumInput(leftMic,iBin).imag()*spectrumInput(rightMic,iBin).imag();
	  float CrossCorrelationI = spectrumInput(leftMic,iBin).imag()*spectrumInput(rightMic,iBin).real() - spectrumInput(leftMic,iBin).real()*spectrumInput(rightMic,iBin).imag();
	  float CC = CrossCorrelationR*CrossCorrelationR + CrossCorrelationI*CrossCorrelationI;
	  float absCC = sqrtf(CC);
	  avgNoisePowerSpectrum[iBin] += absCC;
	  varNoisePowerSpectrum[iBin] += CC;
	}

    } else if (count == numberOfNoiseFrames) {
      for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	avgNoisePowerSpectrum[iBin] /= numberOfNoiseFrames;
      for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++) {
	varNoisePowerSpectrum[iBin] /= numberOfNoiseFrames;
	varNoisePowerSpectrum[iBin] -= avgNoisePowerSpectrum[iBin] ;
      }
    }
    // ----------  Averaging noise spectrum : End ----------



    // ----------  Generalized cross-correlation (GCC) method weighted by phase transform (PHAT) : Begin ----------

    // do signal processing
    if (count > numberOfNoiseFrames) {

      // make steering phase delay
      if (count == numberOfNoiseFrames+1) {

	const float soundSpeed = 34050.f;

	steeringExponentialR = new float*[181];
	steeringExponentialI = new float*[181];
	for (int iDegree = -90; iDegree <= +90; iDegree++) {
	  steeringExponentialR[iDegree+90] = new float[nCols];
	  steeringExponentialI[iDegree+90] = new float[nCols];
	}

	// far-field assumption in a free space environment
	if (diffractionCompensation == "FREE_SPACE")
	  for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	    for (int iDegree = -90; iDegree <= +90; iDegree++) {

	      // Time delay in a free space environment
	      float TimeDelay_free = micDistance*sin(M_PI*float(iDegree)/180.f)/soundSpeed;

	      steeringExponentialR[iDegree+90][iBin] = cos( 2.f*M_PI*float(samplingRate)*float(iBin)/(float(nCols)*2.f) * TimeDelay_free );
	      steeringExponentialI[iDegree+90][iBin] = sin( 2.f*M_PI*float(samplingRate)*float(iBin)/(float(nCols)*2.f) * TimeDelay_free );
	    }

	// far-field assumption with a spearical robot head
	// far-field: distanceBetweenSourceAndMicrophone > ((micDistance^2)*samplingRate/soundSpeed)
	else if (diffractionCompensation == "SPHERICAL_HEAD")
	  for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	    for (int iDegree = -90; iDegree <= +90; iDegree++) {

	      // Time delay for the path around the front of the spearical robot head
	      float TimeDelay_front = micDistance/2.f*(M_PI*float(iDegree)/180.f+sin(M_PI*float(iDegree)/180.f)) / soundSpeed;

	      // Multipath interference created by the two paths around the front and back of the spearical robot head
	      float scallingFactor = 0.1;
	      int signumFunction = 1;	if (iDegree < 0) signumFunction = -1;
	      float TimeDelay_inter = micDistance/2.f* (signumFunction*M_PI - 2.f*M_PI*float(iDegree)/180.f) * abs(sin(M_PI*float(iDegree)/180.f) * scallingFactor) / soundSpeed;

	      // Time delay to compensate for the diffraction of sound waves with multipath interference
	      float TimeDelay_multi = TimeDelay_front - TimeDelay_inter;

	      steeringExponentialR[iDegree+90][iBin] = cos( 2.f*M_PI*float(samplingRate)*float(iBin)/(float(nCols)*2.f) * TimeDelay_multi);
	      steeringExponentialI[iDegree+90][iBin] = sin( 2.f*M_PI*float(samplingRate)*float(iBin)/(float(nCols)*2.f) * TimeDelay_multi);
	    }
      } // if (count == numberOfNoiseFrames+1)

      // do GCC-PHAT
      float sumGccPhatResponseR[181]; // From -90 to +90 degree
      for (int iDegree = -90; iDegree <= +90; iDegree++)
	sumGccPhatResponseR[iDegree+90] = 0.f;

      for (int iBin = frequencyBinFrom; iBin < frequencyBinTo; iBin++)
	{
	  // get data of cross-correlation
	  float CrossCorrelationR = spectrumInput(leftMic,iBin).real()*spectrumInput(rightMic,iBin).real() + spectrumInput(leftMic,iBin).imag()*spectrumInput(rightMic,iBin).imag();
	  float CrossCorrelationI = spectrumInput(leftMic,iBin).imag()*spectrumInput(rightMic,iBin).real() - spectrumInput(leftMic,iBin).real()*spectrumInput(rightMic,iBin).imag();

	  // get a weighting function of the phase transform
	  float CC = CrossCorrelationR*CrossCorrelationR + CrossCorrelationI*CrossCorrelationI;
	  float PHATWeighting = sqrtf(CC);

	  // preserve only interaural phase difference
	  CrossCorrelationR /= PHATWeighting;
	  CrossCorrelationI /= PHATWeighting;

#if 1
	  // get a weighting function based on the Wiener filter and SNR ratios
	  float SNRpost = PHATWeighting / avgNoisePowerSpectrum[iBin]; // calculate a posteriori SNR
	  float SNRinst = SNRpost - 1; // calculate a instantaneous SNR
	  float SNRgain = SNRinst / (1+SNRinst); // get the gain function after appling the Wiener solution
	  if (SNRgain < EPSILON)	SNRgain = EPSILON; // limitation to prevent distortion

	  for (int iDegree = -90; iDegree <= +90; iDegree++)
	    {
	      float gccPhatResponseR = CrossCorrelationR * steeringExponentialR[iDegree+90][iBin] - CrossCorrelationI * steeringExponentialI[iDegree+90][iBin];
	      gccPhatResponseR *= SNRgain; // make gccPhatResponse to be robust to noisy environments
	      sumGccPhatResponseR[iDegree+90] += gccPhatResponseR / nBinsForAveraging;
	    }
#else
	  // get a weighting function based on the statistical SNR ratios (Gaussian function)
	  float StatisticalSNR = CC / varNoisePowerSpectrum[iBin];
	  float GainStatisticalSNR = StatisticalSNR - log(StatisticalSNR) - 1;
	  if (GainStatisticalSNR < EPSILON)	GainStatisticalSNR = EPSILON; // limitation to prevent distortion
	  GainStatisticalSNR = GainStatisticalSNR / (GainStatisticalSNR+1);

	  for (int iDegree = -90; iDegree <= +90; iDegree++)
	    {
	      float gccPhatResponseR = CrossCorrelationR * steeringExponentialR[iDegree+90][iBin] - CrossCorrelationI * steeringExponentialI[iDegree+90][iBin];
	      gccPhatResponseR *= GainStatisticalSNR; // make gccPhatResponse to be robust to noisy environments
	      sumGccPhatResponseR[iDegree+90] += gccPhatResponseR / nBinsForAveraging;
	    }
#endif
	}
      // ----------  Generalized cross-correlation (GCC) method weighted by phase transform (PHAT) : End ----------



      // ----------  Multiple Direction-Of-Arrival (DOA) Estimation (Peaks detection) : Begin ----------
      const int SENSITIVENESS = 0.3;

      // Find derivative
      float differencesBetweenAdjacentIndexes[181+1];
      for (int index = 0; index < 182; index++) {
	if (index == 0)
	  differencesBetweenAdjacentIndexes[index] = 1;
	else if (index == 181)
	  differencesBetweenAdjacentIndexes[index] = -1;
	else {
	  differencesBetweenAdjacentIndexes[index] = sumGccPhatResponseR[index] - sumGccPhatResponseR[index-1];
	  if (differencesBetweenAdjacentIndexes[index] == 0)
	    differencesBetweenAdjacentIndexes[index] = -0.0000000000000000000001;
	}
      }

      // Find where the derivative changes sign
      int numberOfIndexesOfSignChanged = 0;
      bool IndexesOfSignChanged[181];
      for (int index = 0; index < 181; index++) {

	if (differencesBetweenAdjacentIndexes[index]*differencesBetweenAdjacentIndexes[index+1] < 0) {
	  IndexesOfSignChanged[index] = true;
	  numberOfIndexesOfSignChanged++;
	} else
	  IndexesOfSignChanged[index] = false;
      }

      // store where the derivative changes sign
      int* containerOfIndexesOfSignChanged = new int[numberOfIndexesOfSignChanged];
      numberOfIndexesOfSignChanged = 0;
      for (int index = 0; index <  181; index++)
	if (IndexesOfSignChanged[index] == true)
	  containerOfIndexesOfSignChanged[numberOfIndexesOfSignChanged++] = index;

      // process peak detection
      bool* isPeakInContainerOfIndexesOfSignChanged = new bool[numberOfIndexesOfSignChanged];
      for (int index = 0; index < numberOfIndexesOfSignChanged; index++)
	isPeakInContainerOfIndexesOfSignChanged[index] = false;

      if (numberOfIndexesOfSignChanged != 0 && numberOfIndexesOfSignChanged <= 2)
	isPeakInContainerOfIndexesOfSignChanged[0] = true;
      else {

	int candidatePeakIndex = -1000000;
	double candidatePeakValue = -1000000;
	int priorHoleIndex;
	double priorHoleValue;
	int posteriHoleIndex;
	double posteriHoleValue;
	int currentPeakIndex;
	double currentPeakValue;

	for (int index = 0; index < numberOfIndexesOfSignChanged; index += 2) {

	  // adjust value of a prior hole
	  if (index == 0) {
	    priorHoleIndex = index+1;
	    priorHoleValue = sumGccPhatResponseR[containerOfIndexesOfSignChanged[priorHoleIndex]];
	  } else if (priorHoleValue > sumGccPhatResponseR[containerOfIndexesOfSignChanged[index-1]]) {
	    priorHoleIndex = index-1;
	    priorHoleValue = sumGccPhatResponseR[containerOfIndexesOfSignChanged[priorHoleIndex]];
	  }

	  // adjust value of a posteri hole
	  if (index != numberOfIndexesOfSignChanged-1) {
	    posteriHoleIndex = index+1;
	    posteriHoleValue = sumGccPhatResponseR[containerOfIndexesOfSignChanged[posteriHoleIndex]];
	  }

	  // set and adjust value of a peak
	  currentPeakValue = sumGccPhatResponseR[containerOfIndexesOfSignChanged[index]];
	  if (currentPeakValue > candidatePeakValue) {
	    candidatePeakIndex = index;
	    candidatePeakValue = currentPeakValue;
	  }

	  // Peak decision
	  if ( (candidatePeakValue-priorHoleValue) > SENSITIVENESS && (candidatePeakValue-posteriHoleValue) > SENSITIVENESS ) {
	    isPeakInContainerOfIndexesOfSignChanged[candidatePeakIndex] = true;

	    candidatePeakIndex = -1000000;
	    candidatePeakValue = -1000000;
	    priorHoleIndex = posteriHoleIndex;
	    priorHoleValue = posteriHoleValue;
	  }
	} // for (int index = 0; index < numberOfIndexesOfSignChanged; index += 2)
      } // if (numberOfIndexesOfSignChanged == 2)

      // multi-peaks decision with threshold
      int numberOfMultipleDOAs = 0;
      for (int index = 0; index < numberOfIndexesOfSignChanged; index++) {
	if ( isPeakInContainerOfIndexesOfSignChanged[index]==true && sumGccPhatResponseR[containerOfIndexesOfSignChanged[index]] > gccPhatThreshold ) {

	  int estimatedDOA = containerOfIndexesOfSignChanged[index]-90;
	  float powerOfEstimatedDOA = sumGccPhatResponseR[containerOfIndexesOfSignChanged[index]];

	  // Store localization resutls
	  RCPtr<Source> src(new Source());
	  src->id = numberOfMultipleDOAs++;
	  src->power = powerOfEstimatedDOA; // Power of the direction localized
	  src->remaining =  0.f; // An index that assumes how long the localized sound source lasts from then on
	  double theta = estimatedDOA * M_PI / 180.0f;  // angle
	  double phi = 0 * M_PI / 180.0f;   // elevation
	  double cos_phi = cos(phi);
	  src->x[0] = (float)(cos(theta) * cos_phi);
	  src->x[1] = (float)(sin(theta) * cos_phi);
	  src->x[2] = (float)(sin(phi));
	  sourcesPtr->push_back(src);
	  //cout << "SSL:Frame#" << count << ", numberOfMultipleDOAs" << numberOfMultipleDOAs << ", powerOfEstimatedDOA" << powerOfEstimatedDOA << ", estimated DOA: " << estimatedDOA << endl;
	}
      }

      delete [] containerOfIndexesOfSignChanged;
      delete [] isPeakInContainerOfIndexesOfSignChanged;

      // ----------  Multiple Direction-Of-Arrival (DOA) Estimation (Peaks detection) : end ----------


    } // else if (count > numberOfNoiseFrames)
  }
};

